# PbDLib C++ Library

Contributors: Davide De Tommaso, Leonel Rozo, Tohid Alizadeh, Milad Malekzadeh, João Silvério, Sylvain Calinon, Danilo Bruno, Martijn Zeestraten, Ioannis Havoutis and Daniel Berio. 
Pbdlib has been jointly developed at the Italian Institute of Technology, Genoa, Italy and at the Idiap Research Institute, Martigny, Switzerland, under a GPL Version 3 license.

See also https://gitlab.idiap.ch/rli/pbdlib-matlab for a Matlab/GNU-Octave version of the library containing additional experimental functionalities.

### Dependencies

PbDLib uses the Armadillo library for linear algebra, which runs faster if Lapack and OpenBLAS are also installed. 

To install these dependencies, you can run:
```
sudo apt-get install cmake liblapack3 liblapack-dev libopenblas-dev
```

If you don't have C++11 support, you can install it with:
```
sudo add-apt-repository ppa:ubuntu-toolchain-r/test
sudo apt-get update
sudo apt-get install g++-4.9
```

The latest version of Armadillo can then be downloaded from [here](http://arma.sourceforge.net/download.html).
 
### Installation instructions

```
git clone git@gitlab.idiap.ch:rli/pbdlib.git
cd pbdlib
mkdir build
cd build
cmake ..
make
sudo make install
```

A GUI can be used after installation of the library: 
https://gitlab.idiap.ch/rli/pbdlib_gui

### Test 

```
cd examples
./test_gmm
```

### Using PbDLib in your program

In order to use PbDLib in your program, its CMakeLists.txt must be edited to add the include directories of PbDLib and Armadillo, as well as to link it with both libraries.

#### 1) Add PbDLib and Armadillo include directories

The following commands add the include directories:
```
find_package(Armadillo 5.4 REQUIRED)
include_directories(include ${ARMADILLO_INCLUDE_DIRS})
```

The installation of PbDLib copies the header files to /usr/local/include, so the command include_directories(include) 
adds this directory to the search path.

Header files can then be included in source code, by using for example (for the GMM class):
```
#include "pbdlib/gmm.h"   
```

#### 2) Linking your program with PbD and Armadillo

In order to link your program with PbDLib and Armadillo, add the following commands to CMakeLists.txt:
```
find_library(ARMADILLO_LIBRARIES armadillo)
target_link_libraries(yourProgram pbd ${ARMADILLO_LIBRARIES})
```

### References

Did you find PbDlib useful for your research? Please acknowledge the authors in any academic publications that used parts of these codes.
<br><br>

[1] Tutorial (GMM, TP-GMM, MFA, MPPCA, GMR, LWR, GPR, MPC, LQR, trajGMM):
```
@article{Calinon16JIST,
	author="Calinon, S.",
	title="A Tutorial on Task-Parameterized Movement Learning and Retrieval",
	journal="Intelligent Service Robotics",
	publisher="Springer Berlin Heidelberg",
	year="2016",
	volume="9",
	number="1",
	pages="1--29",
	doi="10.1007/s11370-015-0187-9",
}
```

[2] HMM, HSMM:
```
@article{Rozo16Frontiers,
	author="Rozo, L. and Silv\'erio, J. and Calinon, S. and Caldwell, D. G.",
	title="Learning Controllers for Reactive and Proactive Behaviors in Human-Robot Collaboration",
	journal="Frontiers in Robotics and {AI}",
	year="2016",
	month="June",
	volume="3",
	number="30",
	pages="1--11",
	doi="10.3389/frobt.2016.00030"
}
```

[3] Riemannian manifolds (S2,S3):
```
@article{Zeestraten17RAL,
	author="Zeestraten, M. J. A. and Havoutis, I. and Silv\'erio, J. and Calinon, S. and Caldwell, D. G.",
	title="An Approach for Imitation Learning on {R}iemannian Manifolds",
	journal="{IEEE} Robotics and Automation Letters ({RA-L})",
	year="2017",
	month="June",
	volume="2",
	number="3",
	pages="1240--1247"
	doi="10.1109/LRA.2017.2657001",
}
```

[4] Riemannian manifolds (S+):
```
@inproceedings{Jaquier17IROS,
	author="Jaquier, N. and Calinon, S.", 
	title="Gaussian Mixture Regression on Symmetric Positive Definite Matrices Manifolds: Application to Wrist Motion Estimation with {sEMG}",
	booktitle="Proc. {IEEE/RSJ} Intl Conf. on Intelligent Robots and Systems ({IROS})",
	year="2017",
	month="September",
	address="Vancouver, Canada",
	pages=""
}
```

[5] Semi-tied GMM:
```
@article{Tanwani16RAL,
	author="Tanwani, A. K. and Calinon, S.",
	title="Learning Robot Manipulation Tasks with Task-Parameterized Semi-Tied Hidden Semi-{M}arkov Model",
	journal="{IEEE} Robotics and Automation Letters ({RA-L})",
	year="2016",
	month="January",
	volume="1",
	number="1",
	pages="235--242",
	doi="10.1109/LRA.2016.2517825"
}
```

[6] DP-means:
```
@article{Bruno17AURO,
	author="Bruno, D. and Calinon, S. and Caldwell, D. G.",
	title="Learning Autonomous Behaviours for the Body of a Flexible Surgical Robot",
	journal="Autonomous Robots",
	year="2017",
	month="February",
	volume="41",
	number="2",
	pages="333--347",
	doi="10.1007/s10514-016-9544-6"
}
```

[7] Shared control, adaptive teleoperation:
```
@inproceedings{Havoutis17ICRA,
	author="Havoutis, I. and Calinon, S.",
	title="Supervisory teleoperation with online learning and optimal control",
	booktitle=ICRA,
	year="2017",
	month="May-June",
	address="Singapore",
	pages="1534--1540"
}
```

[8] Keypoint-based motion edition through MPC:
```
@inproceedings{Berio17GI,
	author="Berio, D. and Calinon, S. and Fol Leymarie, F.",
	title="Generating Calligraphic Trajectories with Model Predictive Control",
	booktitle="Proc. 43rd Conf. on Graphics Interface",
	year="2017",
	month="May",
	address="Edmonton, AL, Canada",
	pages="132--139",
	doi="10.20380/GI2017.17"
}
```

[9] Adaptive assistance:
```
@article{Pignat17RAS,
	author="Pignat, E. and Calinon, S.",
	title="Learning adaptive dressing assistance from human demonstration",
	journal="Robotics and Autonomous Systems",
	year="2017",
	month="July",
	volume="93",
	number="",
	pages="61--75",
	doi="10.1016/j.robot.2017.03.017",
}
```

### List of examples

This project will build a number of executables in the 'examples' folder, as listed in the table below.

| Filename | ref. | Description |
|----------|------|-------------|
| test_adhsmm.cpp | [2] | Computation of the forward variable of an ADHSMM for a given external input that modifies the duration probabilities of the model |
| test_gmm.cpp | [1] | Gaussian mixture model (GMM) |
| test_gmr*.cpp | [1] | Gaussian mixture regression (GMR) |
| test_hsmm*.cpp | [2] | Hidden semi-Markov model (HSMM) | 
| test_lqr.cpp | [1] | Linear quadratic regulator (LQR) with either finite or infinite horizons, continuous version, iterative computation | 
| test_lqr_discrete.cpp | [1] | Discrete version of LQR |
| test_mpc.cpp | [1] | Model predictive control (MPC), batch LQR computation |
| test_onlineDP.cpp  | [6] | DP-means for online estimation of GMM parameters | 
| test_quaternions.cpp  | [3] | Orientations as unit quaternions |
| test_rewardWeightedRefinement.cpp  | [] | EM-based stochasitc optimization |
| test_tpgmm*.cpp  | [1] | Task-parameterized Gaussian mixture model (TP-GMM) | 
| test_tphsmm.cpp  | [1,2] | Task-parameterized hidden semi-Markov model (TP-HSMM) |
| test_trajGMM.cpp  | [1] | GMM with dynamic features (trajectory GMM) | 
| test_trajMPC*.cpp  | [] | Generation of trajectory distributions with MPC | 


