// Ajay Tanwani, 13 september 2016

#include "pbdlib/gmm.h"
#include "pbdlib/gmr.h"
#include <pbdlib/lqr.h>
#include <pbdlib/quaternion.h>
#include <pbdlib/tpgmm.h>
#include <pbdlib/tphsmm.h>

#include <sstream>
#include <string>
#include "armadillo"
#include <stdio.h>

using namespace pbdlib;
using namespace arma;
using namespace std;

void print_matrix(arma::mat matrix){ cout << matrix << endl;}

int main(int argc, char **argv)
{
	//learning model initializations
	uint nbFrames, nbStates, nbVarPos, nbD, nbVar;
	int nbSamples, demo_id, dataset_num;
	double dt, rf, dt_sm;
	mat A, B, R, A1d, B1d, R_lqr_FH;
	float iFactor, rFactor, lqrFH_factor, g_factor;
	mat AlphaHSMM;
	uint nbVarIn, nbVarOut;

	nbFrames = 2;
	nbStates = 7;
	nbVarPos = 7;
	nbVarIn = 7;
	nbVarOut = 7;
	nbVar = 14; //X,Q, Xo, Qo

	TPHSMM *tphsmm;
	mat rData;
	int pred_length, sampling_method, horizon_method, sim_method;
	uint auto_mode_counter;

	//shared teleop initializations
	TaskParameters *TPs;
	GMM_Model *ProdGauss, *gmmOut, *gmmP;
	std::vector<TaskParameter> cTaskParam, refTaskParam;
	GMR *gmr;
	urowvec in, out;
	rowvec pg_priors;
	mat teleop_sigma;
	float scale_teleop_sigma;
	float offset_hands[3];

	mat Data;

	string loadPath = "data/valve_data_traj.txt";
	Data.load(loadPath, raw_ascii);
	Data = Data.t();
//	cout << "Data: " << Data.t() << endl;

	string saveDir;

	saveDir = "/home/ajay/ros_ws/src/baxter_rli/teleop_ctrl/models/tphsmm/ValveICRA2/";

	string priorsPath = saveDir + "HSMM__priors.txt";
	string varNamesPath = saveDir + "HSMM__varnames.txt";
	string muPrefix = saveDir + "HSMM__mu";
	string SigmaPrefix = saveDir + "HSMM__sigma";

	string transPath = saveDir + "HSMM__trans.txt";
	string durMu = saveDir + "HSMM__durMu.txt";
	string durSigma = saveDir + "HSMM__durSigma.txt";

	stringstream frameFile, index;
	demo_id = 1;
	frameFile.str("");
	index.str("");
	index<<(demo_id);

	frameFile << saveDir << "HSMM__Frame_" << index.str() << ".txt";

	TPs = new TaskParameters(nbVar, nbFrames);
	TPs->loadFromFile(frameFile.str());
	cTaskParam = TPs->getTaskParameters();
	refTaskParam = cTaskParam;

//	for (int j=0;j < nbFrames; j++){
//		for (int k=0; k < nbVar; k++){
//			cout << cTaskParam[j].b[k] << ",\t";
//		}
//		cout << endl;
//	}

//	cout << "Task Param.A" << cTaskParam[0].A << endl;
//	cout << "cTaskParam[0].b" << cTaskParam[0].b << endl;

	// initialize the model
	tphsmm = new TPHSMM(nbVar, nbStates, nbFrames);

	tphsmm->setSamplingMethod(1);
	tphsmm->loadTPHSMMfromFiles(priorsPath,varNamesPath,muPrefix, SigmaPrefix, transPath, durMu, durSigma);

//	cout << tphsmm->getTRANSITION() << endl;
//
//	cout << tphsmm->getDurMU(0) << endl;
//	cout << tphsmm->getDurSIGMA(0) << endl;
//
//	cout << tphsmm->getDurMU(1) << endl;
//	cout << tphsmm->getDurSIGMA(1) << endl;

//	cout << tphsmm->getPRIORS() << endl;
	tphsmm->setCurrentTaskFrame(cTaskParam);
	cTaskParam = tphsmm->getTaskParameters();

	tphsmm->setScalingR((double)iFactor);
	tphsmm->setPredictionLength(pred_length);
	tphsmm->setRegularizationConstLQR((double)rFactor);
	rData = zeros(nbVarIn, pred_length);
	AlphaHSMM = zeros(tphsmm->getNumSTATES(), pred_length);


	ProdGauss = new GMM_Model(nbStates, nbVar);

	for (int i=0; i < nbStates ; i++){
		ProdGauss->setMU(i,cTaskParam[0].A * tphsmm->get_TPGMM()->getGMMS(0).getMU(i) + cTaskParam[0].b );
		ProdGauss->setSIGMA(i,cTaskParam[0].A * tphsmm->get_TPGMM()->getGMMS(0).getSIGMA(i) * cTaskParam[0].A.t());
	}

	pg_priors = rowvec({0.1462, 0.1097, 0.1619, 0.1391, 0.1988, 0.0898, 0.1545});
	ProdGauss->setPRIORS(pg_priors);
//	cout << "tphsmm Priors: " << tphsmm->getPRIORS() << endl;
//	cout << "ProdGauss Priors" << ProdGauss->getPRIORS() << endl;
//	cout << "ProdGauss Number of States " << ProdGauss->getNumSTATES() << endl;

	out  = urowvec({7,8,9,10,11,12,13});
	in  = urowvec({0,1,2,3,4,5,6});

	scale_teleop_sigma = 0.03;
	teleop_sigma = arma::eye(nbVarIn, nbVarIn)*scale_teleop_sigma;

	gmr = new GMR(ProdGauss);
	gmmOut = new GMM_Model(1, nbVarIn);

	//testing GMR out
	gmmOut = gmr->regression(ProdGauss->getMU(0).rows(0,nbVarIn - 1), in, out);
//	cout << "ProdGaussian - Mu 0 " << ProdGauss->getMU(0).t() << endl;
//	cout << "gmmOut - Mu oUT" << gmmOut->getMU(0).t() << endl;
//
//	cout << "ProdGaussian - Sigma 0 " << ProdGauss->getSIGMA(0) << endl;
//	cout << "gmmOut Sigma 0" << gmmOut->getSIGMA(0) << endl;

	gmmOut = gmr->regression(ProdGauss->getMU(1).rows(0,nbVarIn - 1), in, out);
//	cout << "ProdGaussian - Mu 1" << ProdGauss->getMU(1).t() << endl;
//	cout << "gmmOut - Mu oUT" << gmmOut->getMU(0).t() << endl;
//
//	cout << "ProdGaussian - Sigma 1 " << ProdGauss->getSIGMA(1) << endl;
//	cout << "gmmOut Sigma OuT" << gmmOut->getSIGMA(0) << endl;

	gmmP = new GMM_Model(1,nbVarIn);

//	ProdGauss = tphsmm->get_ProdGMM(cTaskParam);
	ProdGauss->setPRIORS(pg_priors);
//	for (int i=0; i < nbStates; i++){
//		ProdGauss->setSIGMA(i,ProdGauss->getSIGMA(i) + eye(nbVar,nbVar)*1E-2);
//	}
//	gmr->setGMMModel(ProdGauss);

////	quat_teleoperator->changeSortConvention(SCALAR_VEC);
//	colvec test_vec;
//	test_vec = colvec({0.4572, 0.1138, 0.1686, 0.1499, 0.1299, -0.7213, 0.6636});
//	gmr->regression(gmmOut, test_vec, in, out);
////			gmr->regression(join_vert(teleoperator_arm_pose.rows(0,2),quat_teleoperator->getAllCoeffs()), in, out);
//
//	// Take product of Gaussians with the teleoperator arm
//	mat SigmaP = (teleop_sigma.i() + gmmOut->getSIGMA(0).i()).i();
//	gmmP->setSIGMA(0,SigmaP);
//	gmmP->setMU(0,SigmaP*(teleop_sigma.i()*test_vec + gmmOut->getSIGMA(0).i()*gmmOut->getMU(0)));
////	quat_teleoperator->changeSortConvention(VEC_SCALAR);
//	vec MuOffset = gmmP->getMU(0);
//
////	cout << "test vector: " << test_vec << endl;
	double regF = 1E-2;

	for (int t=0; t < Data.n_cols; t++){
		mat SigmaTmp = zeros(nbVarOut, nbVarOut);
		colvec MuTmp = zeros(nbVarOut,1);
		for (int m=0; m < nbFrames; m++){
			for (int i=0; i < nbStates ; i++){
//				ProdGauss->setMU(i,cTaskParam[m].A * tphsmm->get_TPGMM()->getGMMS(m).getMU(i) + cTaskParam[m].b );
//				ProdGauss->setSIGMA(i,cTaskParam[m].A * tphsmm->get_TPGMM()->getGMMS(m).getSIGMA(i) * cTaskParam[m].A.t());
				ProdGauss->setMU(i,tphsmm->get_TPGMM()->getGMMS(m).getMU(i));
				ProdGauss->setSIGMA(i,tphsmm->get_TPGMM()->getGMMS(m).getSIGMA(i));
			}
			gmr->setGMMModel(ProdGauss);
			gmr->regression2(gmmOut, Data.col(t), in, out, regF);
			gmmOut->setMU(0,cTaskParam[m].A.submat(nbVarIn,nbVarIn,nbVar-1,nbVar-1)*gmmOut->getMU(0) + cTaskParam[m].b.subvec(nbVarIn,nbVar-1));
			gmmOut->setSIGMA(0, cTaskParam[m].A.submat(nbVarIn,nbVarIn,nbVar-1,nbVar-1)*gmmOut->getSIGMA(0)*cTaskParam[m].A.submat(nbVarIn,nbVarIn,nbVar-1,nbVar-1).t());

			SigmaTmp += gmmOut->getSIGMA(0).i();
			MuTmp += gmmOut->getSIGMA(0).i()*gmmOut->getMU(0);
//			cout << gmmOut->getMU(0).t();
		}
		gmmOut->setMU(0,SigmaTmp.i()*MuTmp);
		gmmOut->setSIGMA(0,SigmaTmp.i());

		// Take product of Gaussians with the teleoperator arm
		mat SigmaP = (teleop_sigma.i() + gmmOut->getSIGMA(0).i()).i();
		gmmP->setSIGMA(0,SigmaP);
		gmmP->setMU(0,SigmaP*(teleop_sigma.i()*Data.col(t) + gmmOut->getSIGMA(0).i()*gmmOut->getMU(0)));

		cout << gmmOut->getMU(0).t() << endl;
	}

	return 0;
}
