// Author: Ajay Tanwani, 2016

//Discrete lqr tests for finite and infinite horizon
#include <iostream>
#include <sstream>
#include "pbdlib/lqr.h"
#include "armadillo"
#include <vector>


using namespace pbdlib;
using namespace arma;

void print_matrix(arma::mat matrix){ cout << matrix << endl;}


int Factorial(int x, int result = 1) {

  if (x == 1 || x == 0){
	  return result;
  }else{
	  return Factorial(x - 1, x * result);
  }
}

mat diagMat(vec d, int k){

	mat dout = zeros(d.n_rows + abs(k),d.n_rows + abs(k));

	for (int i=0; i < d.n_rows; i++){
		dout(i,i+abs(k)) = d.at(i);
	}
	return dout;
}

int main(int argc, char **argv)
{
	int nbVar = 2;
	int nbDeriv = 2;  //for [x,dx];
	mat A,B;
	float dt,rFactor;
	int nbData = 10;
	mat R,Target;
	std::vector<float> DataIn;
	std::vector<mat> Q;

	dt = 0.01;
	rFactor = 1E-5;

	mat A1d = zeros(nbDeriv,nbDeriv);
	mat B1d = zeros(nbDeriv,1);

	for (int i=0; i < nbDeriv; i++){
		A1d = A1d + diagMat(ones(nbDeriv-i,1),i)* pow(dt,i) * 1/Factorial(i);
	}

	for (int i=0; i < nbDeriv; i++){
		B1d.at(nbDeriv - 1 - i) = pow(dt,i+1) * 1/Factorial(i+1);
	}

	A = kron(A1d, eye(nbVar,nbVar));
	B = kron(B1d, eye(nbVar,nbVar));

	R = eye(nbVar,nbVar)*rFactor;

	Target = zeros(nbVar*2,nbData);

	Target.row(0) = 100*ones(1,nbData);
	Target.row(1) = 50*ones(1,nbData);

	for (int t=0;t<nbData;t++){
		DataIn.push_back(t*dt);
		mat QTmp = zeros(nbVar*2,nbVar*2);
//		QTmp(0,0) = 1;
		QTmp.submat(0,0,nbVar-1,nbVar-1) = eye(nbVar,nbVar);
		Q.push_back(QTmp);
	}

	LQR test_lqr(A,B,dt);

	test_lqr.setProblem(R,Q,Target);

	//Test infinite horizon
	test_lqr.evaluate_gains_infiniteHorizon_Discrete();

	//Test finite horizon
//	mat S;
//	colvec d;
//
//	S = zeros(nbVar*2,nbVar*2);
//	d = zeros(nbVar*2,1);
//
//	test_lqr.evaluate_gains_finiteHorizon(S,d);

	//Output data
	for (int t=0;t<nbData;t++){
		std::cout << t << " , " << test_lqr.getGains().at(t) << std::endl;
	}
	std::cout << test_lqr.getFF() << std::endl;

	return 0;
}
