/*
 * test_tphsmm.cpp
 *
 *  Created on: Sep 17, 2015
 *      Author: Ajay Tanwani
 */

#include <pbdlib/tphsmm.h>

using namespace arma;
using namespace pbdlib;
using namespace std;

int main(int argc, char **argv)
{
	// Model variables
	uint nbFrames = 2;
	uint nbStates = 7;
	uint nbVarPos = 7;
	uint nbVar = 14; //x, y, z, qw, qx, qy, qz, \dot{x}, \dot{y}, \dot{z}, \dot{qw}, \dot{qx}, \dot{qy}, \dot{qz}
	uint nbD = 200; //Number of datapoints in a demonstration
	uint nbSamples = 8;
	double dt = 0.1;
	uint train_samples_ids[4] = {1, 2, 3, 4};
	uint test_samples_ids[4] = {5,6,7,8};

	const char* varNames[] = {"x", "y", "z", "qw", "qx", "qy", "qz", "x_dot", "y_dot", "z_dot", "qw_dot", "qx_dot", "qy_dot", "qz_dot"};
//	varNames.push_back("t");

	TPHSMM *myTphsmm;
//	myTphsmm = new TPHSMM(nbVar, nbStates, nbFrames);
	myTphsmm = new TPHSMM(nbVar, nbStates, nbFrames, nbVarPos, dt, 1E1);


	string priorsPath = "./data/tphsmm/HSMM_check_priors.txt";
	string varNamesPath = "./data/tphsmm/HSMM_check_varnames.txt";
	string muPrefix = "./data/tphsmm/HSMM_check_mu";
	string SigmaPrefix = "./data/tphsmm/HSMM_check_sigma";

	string transPath = "./data/tphsmm/HSMM_check_trans.txt";
	string durMu = "./data/tphsmm/HSMM_check_durMu.txt";
	string durSigma = "./data/tphsmm/HSMM_check_durSigma.txt";

	// Load Var names from file.
	std::ifstream varsfile(varNamesPath.c_str()); // Open file stream
//	std::vector<std::string> vars; 			   // Vector to hold variable names
	std::string varsnames;                     // tmp string to read line from file
//	std::string buf; 						   // buffer string
	// We assume that the variable names are on the first line, separated by spaces
  	getline(varsfile,varsnames);
  	cout << varsnames << endl;

	myTphsmm->loadTPHSMMfromFiles(priorsPath, varNamesPath, muPrefix, SigmaPrefix, transPath, durMu, durSigma);

	cout << "Number of States: " << myTphsmm->getNumSTATES() << endl;
	//..................................................................................


	//......... loading the task parameters from the txt files ......
	string framePrefix = "./data/tphsmm/HSMM_check_Frame_";
	stringstream frameFile, index;

	std::vector<TaskParameter> cTaskParam;
	mat rData;
	colvec init_state;

	//give me the training sample frames
	for (int i=0; i < 1; i++){
		frameFile.str("");
		index.str("");
		index<<(train_samples_ids[i]);

		frameFile    << framePrefix << index.str() << ".txt";
		cout << "Frame 1" << endl;
		myTphsmm->setCurrentTaskFrameFromFile(frameFile.str());
		cout << "Frame 2" << endl;
		cTaskParam = myTphsmm->getTaskParameters();
		cout << "Frame 3" << endl;
		rData = myTphsmm->TPHSMMController();
		for (int j=0; j < rData.n_rows; j++){
			for (int k=0; k < rData.n_cols; k++){
				cout << rData.at(j,k) << ",\t";
			}
			cout << endl;
		}
//		for (int j=0;j < nbFrames; j++){
//			for (int k=0; k < nbVar; k++){
//				cout << cTaskParam[j].b[k] << ",\t";
//			}
//			cout << endl;
//		}

//		myTphsmm->TPHSMMController();


	}

	//give me the test sample frames

	//generate random frames by interpolation or some other method

	//take product of Gaussians for the current frame

	//generate sample sequence by deterministic and/or stochastic interpolation from the current state for the current frame

	//Update current state

	//Add noise to the current state - mean 0, std, n1

	//Add noist to every state - mean 0, std n2

//	std::vector<TPDemonstration> demos; // Vector to hold demonstrations
//	TPDemonstration tmpDemo; // Temp variable to load individual demonstrations
//
//	char Datafilename[256];
//	char TPfilename[256];
//	cout << "Loading the demonstrations and the task parameters ..." << endl;
//	for (uint m=0; m<nDemos; m++){   //Loading demos in the loop
//		// Form filenames:
//		sprintf(Datafilename, "../data/pgmm/Data0%d.txt",m+1);
//		sprintf(TPfilename, "../data/pgmm/Param0%d.txt",m+1);
//
//		// Load files in demo:
//		// note we need to transpose the data (last argument)
//		// because the data in the file a row for each data point
//		// and a column for each variable. The agreement is
//		// the other way around
//		tmpDemo.loadFromFiles(Datafilename,TPfilename, true);
//
//		// push back demonstration:
//		demos.push_back(tmpDemo);
//	}
//
//	cout << "Demonstrations Loaded succesfully." << endl;
//	cout<<"Press any key to continue..."<<endl; getchar();
//
//	//..................................................................................
//	//......... Learning the PGMM model parameters using the EM algorithm ......
//	cout<<"Learning model parameters using EM..." << endl;
//
//	TPGMM *tpgmm;
//	tpgmm = new TPGMM(demos, nStates);
//
//	cout << "\n Number of EM iterations: " << tpgmm->estimateTensorGMM() << endl;
//
//	// saving the learned model parameters
//	mat MuTmp = zeros(nVars, nStates*nFrames);
//	mat SigmaTmp = zeros(nVars, nVars*nStates*nFrames);
//	mat Priors = zeros(1,nStates);
//	std::vector<GMM_Model> GMMSresult;
//	GMMSresult = tpgmm->getGMMS();
//	for(uint m=0; m<nFrames; m++){
//		for (uint i=0;i<nStates; i++){
//			MuTmp.col(m*nStates+i) = GMMSresult[m].getMU(i);
//			SigmaTmp.cols(m*nVars*nStates+i*nVars, m*nVars*nStates+i*nVars+nVars-1) = GMMSresult[m].getSIGMA(i);
//		}
//	}
//	sprintf(Datafilename, "../../data/pgmm/Mu0%d.txt",2);
//	MuTmp.save(Datafilename, raw_ascii);
//	sprintf(Datafilename, "../../data/pgmm/Sigma0%d.txt",2);
//	SigmaTmp.save(Datafilename, raw_ascii);
//	sprintf(Datafilename, "../../data/pgmm/Priors0%d.txt",2);
//	Priors = tpgmm->getPRIORS();
//	Priors.save(Datafilename, raw_ascii);
//	cout<<"the learned model parameters are written into text files succesfully."<<endl; cout<<"Press any key to continue..."<<endl; getchar();
//
//
//	// Reproduction variables (GMR)
//	GMR	*gmr = new GMR();
//	Datapoints	*repro = new Datapoints(1,1);  // Only one datapoint passed at each time step
//	mat	reproData, auxYgmr;
//	std::vector<std::string> reproVarNames;
//	reproVarNames.push_back("t");              // time is the query point
//	repro->setVarNames(reproVarNames);
//
//	// loading new parameters for the reproduction from the text file
//	// In this case we have considered that the parameters are the same all along the reproduction. They can be varying.
//	TaskParameters ReproTPs(nVars,nFrames);
//	sprintf(TPfilename, "../../data/pgmm/ParamRepro0%d.txt",1);
//	ReproTPs.loadFromFile(TPfilename);
//
//	GMM_Model* gmm;
//	for(uint tn=1; tn<=nData; tn++){
//		// Computing the resulting GMM given the set of parameters {A,b}
//		gmm = tpgmm->getTransformedGMM(ReproTPs);
//		gmm->setVARSNames(varNames);
//
//		// Printing model components
//		cout << "Resulting GMM given the set of parameters 'A' and 'b'" << endl;
//		cout << "nbStates : " << gmm->getNumSTATES() << endl;
//		cout << "nbVar    : " << gmm->getNumVARS() << endl;
//		cout << "VarNames " ;
//		for (uint i=0;i<gmm->getNumVARS();i++)
//			cout << varNames[i] ;
//		cout << endl;
//
//		cout << "Priors   : "  << gmm->getPRIORS() << endl;
//		for(uint i=0; i<nStates; i++){
//			cout << "State #" << i << ":"<< endl;
//			gmm->getMU(i).print("Mu = ");
//			gmm->getSIGMA(i).print("Sigma = ");
//		}
//
//		// Computing GMR
//		gmr->setGMMModel(gmm);
//		mat repDataPt;
//		repDataPt << (tn * 0.01);	// time step as input
//		repDataPt.print("tn = ");
//		repro->setData(repDataPt);
//		GMM_Model* gmmOut = gmr->regression(repro);
//		gmmOut->getMU(0).print("MuOut = ");
//
//		cout << "Please press [ENTER] to continue or [CTRL+C] to finish" << endl;
//		char key;
//		std::cin.ignore(1);
//	}
	return 0;
}



