/* Martijn Zeestraten, December 2015
 *
 * Demo shows how the TrajMPC class can be used to generate trajectory distributions
 *
 */

#include <time.h>
#include "armadillo"
// Pbdlib dependencies:
#include "pbdlib/hsmm.h"
#include "pbdlib/taskparameters.h"
#include "pbdlib/gmm.h"
#include "pbdlib/trajMPC.h"
#include "pbdlib/trajdist.h"


using namespace arma;
using namespace pbdlib;
using namespace std;


int main(int argc, char* argv[])
{
		
cout << "Setting Variables: ";

// Model Settings:
uint nVarsPos = 2; // Number of position variables 
uint nDeriv   = 2; // Number of derivatives considered

// TrajMPC Settings:
uint Np, Nc; // These values are set later based on the average duration
double dt     = 0.01;  // time step considered
double alpha  = 0.001; // control cost variable

// Initial state of the system:
colvec Xinit;
Xinit << -27.0 << 42.0<< 0.0 << 0.0;

string prefix = "HSMM_test";

// Source Files for the HSMM Model:
string muPath        = "data/hsmm/" + prefix + "_mu.txt";
string sigmaPath     = "data/hsmm/" + prefix + "_sigma.txt";
string priorsPath    = "data/hsmm/" + prefix + "_priors.txt";
string transPath     = "data/hsmm/" + prefix + "_trans.txt";
string durMu         = "data/hsmm/" + prefix + "_durMu.txt";
string durSigma      = "data/hsmm/" + prefix + "_durSigma.txt";

// Create pbdlib objects:
cout << "Create pbdlib objects ";
HSMM*    myHSMM;        // HSMM to handle state sequences
TrajMPC* myTrajMPC;   
cout << "done " << endl;

// Load HSMM:
cout << "load HSMM ";
myHSMM = new HSMM(priorsPath,muPath,sigmaPath,transPath,durMu,durSigma);
cout << "done " << endl;
cout << "HSMM nbStates: " << myHSMM->getNumSTATES() << endl;
cout << "HSMM nbVars  : " << myHSMM->getNumVARS() << endl;

// We want to generate a trajectory distribution for a trajectory of average duration:
Np = 0;
for (int i=0;i<myHSMM->getNumSTATES();i++)
	Np = Np +  myHSMM->getDurMU(i)[0];
Nc = Np;

// Load TrajMPC:
cout << "load TrajMPC ";
myTrajMPC = new TrajMPC(myHSMM,dt,nVarsPos,nDeriv,Np,Nc,alpha);
cout << "done " << endl;
cout << "TrajMPC nbVars  : " << myTrajMPC->getNumVARS() << endl;

// **********************************************************************

// Make sure that the TrajMPC model is correclty initialized by resetting it before starting
myTrajMPC->reset();


//colvec curPos = myTrajMPC->getCurrentAttractor();
colvec curPos = Xinit;
TrajDist TD;
cout << "Current Attractor: " << endl <<curPos<< endl;

int start = clock();
cout << "Simulation of the reproduction: " << endl;
TD = myTrajMPC->computeTrajDist(curPos,false);
int end = clock();
std::cout << "it took " << end - start << " ticks, or " << ((float)end - start)/CLOCKS_PER_SEC << " seconds." << std::endl;


cout << "Writing algorithm output to files... " ;
TD.Sigma.save("trajDist_Sigma.txt",csv_ascii);
TD.Mu.save("trajDist_Mu.txt",csv_ascii);
cout << "Done" << endl;

}

